/// Functions for performing cryptographic operations on the main data structures.

use super::*;
use openssl::{
    symm::{Cipher, Crypter, Mode}
};

/// Errors that can occur during cryptographic operations.
pub(crate) enum Error {
    NoReadCap
}

type Result<T> = std::result::Result<T, Error>;

#[allow(dead_code)]
pub(crate) fn encrypt_block(
    versioned_block: VersionedBlock, principal: &Principal, key: &Key
) -> Result<VersionedBlock> {
    let VersionedBlock::V0(mut block) = versioned_block;
    let mut body = match block.body {
        Cryptotext::Cipher(_) => return Ok(VersionedBlock::V0(block)),
        Cryptotext::Plain(body) => body
    };
    let (principal_owned, read_cap) = block.read_caps.remove_entry(principal)
        .ok_or(Error::NoReadCap)?;
    let block_key = decrypt(read_cap, key)?;
    encrypt_in_place(&mut body, &block_key)?;
    block.body = Cryptotext::Cipher(body);
    block.read_caps.insert(principal_owned, encrypt(block_key, key)?);
    Ok(VersionedBlock::V0(block))
}

#[allow(dead_code)]
pub(crate) fn decrypt_block(
    versioned_block: VersionedBlock, principal: &Principal, key: &Key
) -> Result<VersionedBlock> {
    let VersionedBlock::V0(mut block) = versioned_block;
    let mut body = match block.body {
        Cryptotext::Plain(_) => return Ok(VersionedBlock::V0(block)),
        Cryptotext::Cipher(body) => body
    };
    let (principal_owned, read_cap) = block.read_caps.remove_entry(principal)
        .ok_or(Error::NoReadCap)?;
    let block_key = decrypt(read_cap, key)?;
    decrypt_in_place(&mut body, &block_key)?;
    block.body = Cryptotext::Plain(body);
    block.read_caps.insert(principal_owned, Cryptotext::Plain(block_key));
    Ok(VersionedBlock::V0(block))
}

fn encrypt<T: AsRef<[u8]> + AsMut<[u8]>>(mut data: T, key: &Key) -> Result<Cryptotext<T>> {
    encrypt_in_place(data.as_mut(), key)?;
    Ok(Cryptotext::Cipher(data))
}

fn decrypt<T: AsRef<[u8]> + AsMut<[u8]>>(cryptotext: Cryptotext<T>, key: &Key) -> Result<T> {
    let mut data = match cryptotext {
        Cryptotext::Plain(data) => return Ok(data),
        Cryptotext::Cipher(data) => data
    };
    decrypt_in_place(data.as_mut(), key)?;
    Ok(data)
}

fn encrypt_in_place(_plaintext: &mut [u8], _key: &Key) -> Result<()> {
    unimplemented!();
}

fn decrypt_in_place(_ciphertext: &mut [u8], _key: &Key) -> Result<()> {
    unimplemented!();
}

/// Tests that validate the dependencies of this module.
#[cfg(test)]
mod dependency_tests {
    use super::*;

    static KEY: [u8; 32] = [
        0xB2, 0xB3, 0xDA, 0x5A, 0x1A, 0xF6, 0xB3, 0x78, 0x30, 0xAB, 0x1D, 0x33, 0x33, 0xE7, 0xE3,
        0x5B, 0xBB, 0xF9, 0xFE, 0xD0, 0xC1, 0xF7, 0x90, 0x34, 0x69, 0xB7, 0xE7, 0xC6, 0x1C, 0x46,
        0x85, 0x48,
    ];

    static IV: [u8; 16] = [
        0x60, 0xE0, 0xBE, 0x45, 0xA9, 0xAB, 0x5D, 0x99, 0x3B, 0x3A, 0x1E, 0x54, 0x18, 0xE0, 0x46,
        0xDE,
    ];

    /// This test validates that data decrypted with the `Crypter` API matches data that was
    /// previously encrypted using it.
    #[test]
    fn cryptor() {
        let expected = b"We attack at the crack of noon!";

        let cipher = Cipher::aes_256_cbc();
        let mut encrypter = Crypter::new(
            cipher,
            Mode::Encrypt,
            &KEY,
            Some(&IV)).unwrap();
        let mut ciphertext = vec![0; expected.len() + cipher.block_size()];
        let mut written = encrypter.update(expected.as_slice(), &mut ciphertext).unwrap();
        written += encrypter.finalize(&mut ciphertext[written..]).unwrap();
        ciphertext.truncate(written);

        let mut actual = vec![0u8; ciphertext.len() + cipher.block_size()];
        let mut decrypter = Crypter::new(
            cipher,
            Mode::Decrypt,
            &KEY,
            Some(&IV)).unwrap();
        let mut written = decrypter.update(&ciphertext, &mut actual).unwrap();
        written += decrypter.finalize(&mut actual[written..]).unwrap();
        actual.truncate(written);

        assert_eq!(expected, actual.as_slice());
    }
}

/// Tests for the code that is actually in this module.
#[cfg(test)]
mod tests {

}